import functools as ft
from typing import Dict, List, NamedTuple, Optional, Union, Sequence, Iterable
import itertools
import math


class MergerRule(NamedTuple):
    into: List[int]
    ordering: List[int]


class Merger(NamedTuple):
    mergerules: Dict[int, MergerRule]
    completions: Dict[int, List[int]]
    mergelens: List[int]


class Root(NamedTuple):
    id: int
    head: int


class MergeTree(NamedTuple):
    head: int
    children: Union[List["MergeTree"], Root]
    isnew: bool


def combinations(seq: Sequence[int]):
    return itertools.chain.from_iterable(
        itertools.combinations(seq, i) for i in range(len(seq))
    )


def muticombinations(
    iterable: Sequence[int], lens: Optional[List[int]] = None, min_groups: int = 0
):
    """Compute all possible combinations of a `set` with multiple combinations
    allowed.

    That is, for the set [1,2,3,4], we allow [(1,2),(3,4)]. An optional
    arguments `lens` can be passed to specify the maximum combination length. If
    `lens` is [2,3], then only combinations of length 2 and 3 will be returned.
    """
    combs = [[[]]]
    s = set(iterable)

    for com in combinations(iterable):
        com = list(com)
        if lens and not (len(com) in lens):
            continue

        remaining = list(s.symmetric_difference(com))
        if min_groups < 2:
            new_comb = [com, *[[e] for e in remaining]]
            combs.append(new_comb)

        sub_combs = muticombinations(remaining, lens=lens, min_groups=0)

        for sub_comb in sub_combs:
            n = sum(map(lambda c: len(c) > 1, sub_comb))
            if n + 1 > min_groups:
                combs.append([[c for c in com], *sub_comb])

    return set(map(lambda el: sorted(el, key=lambda a: min(a)), combs))


def make_muticombinations(n: int, merger: Merger):
    return muticombinations(range(n), lens=merger.mergelens, min_groups=2)


def prod(iterable: Iterable):
    return ft.reduce(lambda x, y: x * y, iterable)


def canmerge(merger: Merger, heads: Sequence[int]):
    return merger.mergerules.get(prod(heads)) is not None


def count_new(trees: Sequence[MergeTree]):
    return len(list(itertools.filterfalse(lambda t: t.isnew, trees)))


def heads(trees: Sequence[MergeTree]):
    return map(lambda t: t.head, trees)


def all_have_n_new(groups, n):
    """Determine if every group in groups has at least `n` new trees  (ignoring
    groups with only one tree.)
    """
    return all(
        filter(lambda t: count_new(t) >= n, filter(lambda g: len(g) > 1, groups))
    )


def getperm(seq1: Sequence, seq2: Sequence):
    s1 = sorted(enumerate(seq1), key=lambda t: t[1])
    s2 = sorted(enumerate(seq2), key=lambda t: t[1])
    return [t[1] for t in sorted([(t1[0], t2[0]) for t1, t2 in zip(s1, s2)])]


def mergeone(merged: List[List[MergeTree]], trees: List[MergeTree], merger: Merger):
    """
    Given a set of tree, find all possible ways to merge them to form a new
    tree, add the results to the `merged` array and return `true` to indicate
    the grouping was successful. If the merge failed, return false.
    """
    if len(trees) == 1:
        pass

    hs = tuple(heads(trees))
    if not canmerge(merger, hs):
        return None

    rule = merger.mergerules[prod(hs)]
    perm = getperm(rule.ordering, hs)
    sub = [trees[p] for p in perm]
    merged.append([MergeTree(h, sub, True) for h in rule.into])
    return merged


def mergeall(trees: List[MergeTree], merger: Merger):
    """Merge em"""
    ntrees = len(trees)
    if ntrees == 1:
        return [trees]

    new_trees: List[List[MergeTree]] = []

    for comb in make_muticombinations(ntrees, merger):
        groups = [[trees[i] for i in c] for c in comb]
        if all_have_n_new(groups, 1):
            good = True
            m = len(groups)
            cnt = 1
            while good and cnt <= m:
                group = groups.pop()
                good = mergeone(groups, group, merger)
                cnt += 1

            if good:
                new_trees = unfold_into(new_trees, groups)

    return new_trees


def unfold_into(newtrees: List[List[MergeTree]], groups: List[List[MergeTree]]):
    """Flatten out the trees and add them into the array. Currently, the trees
    have entries like [...([T₁,...,Tₙ])...] where the different T's come from
    different merges. We want to flatten this out so we get all possible
    mergings
    """
    m = prod(map(len, groups))
    ngroups = len(groups)
    for i in range(m):
        freq = 1
        tree: List[MergeTree] = []
        for j in range(ngroups):
            length = len(groups[j])
            freq *= length
            tree[j] = groups[j][math.floor(freq * (i - 1) / m) % length + 1]
        newtrees.append(tree)

    return newtrees


def build_trees(roots: List[Root], merger: Merger):
    """Given a set of roots, turn them into `MTree`s and continue to merge the
    trees into fewer trees using the `mergerules`, which is a dictionary maping
    tuples of heads to a new head: mergerules = Dict([(A,B,C) -> (D,E), ...]).
    After each merging, check if there is a way to `complete` the tree by
    combining all subtrees into a single tree using the `completions` array
    which consists of a vector of tuples specifying which sets of trees can be
    completed: completions = [(A,B,C,D), (A,B,C,E), ...]  Then return all the
    completed trees.
    """
    completed: List[List[MergeTree]] = [[]]
    incomplete: List[List[MergeTree]] = [[]]
    incomplete.append([MergeTree(root.head, root, True) for root in roots])

    while len(incomplete) > 0:
        trees = incomplete.pop()
        hs = tuple(heads(trees))
        key = prod(hs)
        if count_new(trees) >= 2 and merger.completions.get(key):
            perm = getperm(merger.completions[key], hs)
            trees = [trees[p] for p in perm]
            completed.append(trees)

        merged = mergeall(trees, merger)
        if len(merged) > 0:
            incomplete.extend(merged)

    return completed



from typing import Union

import numpy as np
import numpy.typing as npt

Real = Union[np.uint, np.float_, np.complex_, np.int_, np.uint]
Complex = np.complex_
Numeric = Union[Real, Complex]
RealArray = npt.NDArray[Real]
ComplexArray = npt.NDArray[Real]
NumericArray = npt.NDArray[Numeric]



import dataclasses
import enum

import numpy as np

from helax.numpy.typing import ComplexArray, RealArray


class WeylType(enum.IntFlag):
    """Enumeration of the different types of Weyl spinors.

    Members
    -------
    X:
        Incoming left-handed (1/2, 0) fermion.
    Xd:
        Outgoing left-handed (1/2, 0) fermion.
    Y:
        Outgoing right-handed (0, 1/2) fermion.
    Yd:
        Incoming right-handed (0, 1/2) fermion.
    """

    X = enum.auto()
    Xd = enum.auto()
    Y = enum.auto()
    Yd = enum.auto()


@dataclasses.dataclass
class WeylWf:
    """Dataclass for Weyl wavefunctions (2-component spinors).

    Attributes
    ----------
    wavefunction: complex array
        2-Component Dirac spinor.
    momentum: array
        Four-momentum of the spinor. The momentum always points in the
        direction of the fermion flow.
    type: WeylType
        Type of weyl spinor (x, y, x^+, or y^+).
    """

    wavefunction: ComplexArray
    momentum: RealArray
    type: WeylType


# =============================================================================
# ---- Dirac Wavefunction -----------------------------------------------------
# =============================================================================


def __check_spin_dirac(*, spin: int):
    assert spin == 1 or spin == -1, "Spin must be 1 or -1."


def __chi(*, momentum: RealArray, spin: int) -> ComplexArray:
    """
    Compute the two-component weyl spinor.

    Parameters
    ----------
    p: array
        Array containing the 4-momentum of the wavefunction.
    s: int
        Spin of the wavefunction. Must be 1 or -1.
    """
    dtype = momentum.dtype
    eps = np.finfo(dtype.name).eps

    px = momentum[1]
    py = momentum[2]
    pz = momentum[3]

    pm = np.linalg.norm(momentum[1:], axis=0)
    mask = pm + pz <= eps
    x = np.zeros((2, momentum.shape[-1]), dtype=dtype) * 1j

    if np.any(mask):
        # x[0, mask] = 0.0j  # (spin - 1.0) / 2.0 + 0j
        x[1, mask] = spin  # (spin + 1.0) / 2.0 + 0j

    mask = ~mask
    if np.any(mask):
        px = px[mask]
        py = py[mask]
        pz = pz[mask]
        pm = pm[mask]

        den = np.sqrt(2 * pm * (pm + pz))
        x[0, mask] = (pm + pz) / den
        x[1, mask] = (spin * px + py * 1j) / den

    if spin == -1:
        return np.array([x[1], x[0]])
    return np.array([x[0], x[1]])


def __weyl_spinor(
    *, momentum: RealArray, mass: float, spin: int, type: WeylType
) -> ComplexArray:
    """
    Compute the dirac wavefunction.

    Parameters
    ----------
    p:
        Four-momentum of the wavefunction.
    mass:
        Mass of the wavefunction.
    s: int
        Spin of the wavefunction. Must be 1 or -1.
    anti: Int
        If anti = -1, the wavefunction represents a v-spinor.
        If 1, wavefunction represents a u-spinor.
    """
    pm = np.linalg.norm(momentum[1:], axis=0)

    wp = np.sqrt(momentum[0] + pm)
    wm = mass / wp

    if type == WeylType.X:
        x = __chi(momentum=momentum, spin=spin)
        w = wp if spin == -1 else wm

    elif type == WeylType.Yd:
        x = __chi(momentum=momentum, spin=spin)
        w = wm if spin == -1 else wp
        x = np.array([-x[1], x[0]])

    elif type == WeylType.Y:
        x = __chi(momentum=momentum, spin=-spin)
        w = -wp if spin == 1 else wm

    elif type == WeylType.Xd:
        x = __chi(momentum=momentum, spin=-spin)
        x = np.array([-x[1], x[0]])
        w = wm if spin == 1 else -wp

    else:
        raise ValueError(f"Invalid weyl spinor type {type}")

    return w * x


def weyl_x(momentum: RealArray, mass: float, spin: int) -> WeylWf:
    """
    Compute a x-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
        If 2-dimensional, 2nd dimension must be the batch dimension.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    __check_spin_dirac(spin=spin)
    wf = __weyl_spinor(momentum=momentum, mass=mass, spin=spin, type=WeylType.X)
    return WeylWf(
        wavefunction=wf,
        momentum=momentum,
        type=WeylType.X,
    )


def weyl_y(momentum: RealArray, mass: float, spin: int) -> WeylWf:
    """
    Compute a y-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    __check_spin_dirac(spin=spin)
    wf = __weyl_spinor(momentum=momentum, mass=mass, spin=spin, type=WeylType.Y)
    return WeylWf(
        wavefunction=wf,
        momentum=-momentum,
        type=WeylType.Y,
    )


def weyl_xd(momentum: RealArray, mass: float, spin: int) -> WeylWf:
    """
    Compute a x^+-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    __check_spin_dirac(spin=spin)
    wf = __weyl_spinor(momentum=momentum, mass=mass, spin=spin, type=WeylType.Xd)
    return WeylWf(
        wavefunction=wf,
        momentum=momentum,
        type=WeylType.Xd,
    )


def weyl_yd(momentum: RealArray, mass: float, spin: int) -> WeylWf:
    """
    Compute a y^+-spinor wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be 1 or -1.
    """
    __check_spin_dirac(spin=spin)
    wf = __weyl_spinor(momentum=momentum, mass=mass, spin=spin, type=WeylType.Yd)
    return WeylWf(
        wavefunction=wf,
        momentum=-momentum,
        type=WeylType.Yd,
    )


def weyl_raise_index(chi: WeylWf) -> WeylWf:
    """Raise the index of the weyl spinor."""

    # z^a = eps^ab w_b
    #   z^1 = eps^12 w_2 = +w_2
    #   z^2 = eps^21 w_1 = -w_1
    wf = np.array(
        [
            chi.wavefunction[1],
            -chi.wavefunction[0],
        ]
    )
    return WeylWf(wavefunction=wf, momentum=chi.momentum, type=chi.type)


def weyl_lower_index(chi: WeylWf) -> WeylWf:
    """Lower the index of the weyl spinor."""

    # z_a = eps_ab w^b
    #   z_1 = eps_12 w^2 = -w_2
    #   z_2 = eps_21 w^1 = +w_1
    wf = np.array(
        [
            -chi.wavefunction[1],
            chi.wavefunction[0],
        ]
    )
    return WeylWf(wavefunction=wf, momentum=chi.momentum, type=chi.type)



import dataclasses

import numpy as np

from helax.numpy.typing import ComplexArray, RealArray


@dataclasses.dataclass
class ScalarWf:
    wavefunction: ComplexArray
    momentum: RealArray
    direction: int


# =============================================================================
# ---- Scalar Wavefunction ----------------------------------------------------
# =============================================================================


def scalar_wf(momentum: RealArray, out: bool) -> ScalarWf:
    """
    Compute a vector wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    """
    s = -1 if out else 1
    if len(momentum.shape) > 1:
        shape = (momentum.shape[-1],)
    else:
        shape = (1,)

    wf = np.ones(shape, dtype=momentum.dtype) + 0.0j
    return ScalarWf(wavefunction=wf, momentum=s * momentum, direction=s)



import pathlib

import numpy as np
import pytest
from pytest import approx

from helax.numpy.lvector import ldot
from helax.numpy.wavefunctions import vector_wf

TEST_DATA = np.load(
    pathlib.Path(__file__).parent.joinpath("testdata").joinpath("vector_wf_data.npz")
)

MetricG = np.array(
    [
        [1.0, 0.0, 0.0, 0.0],
        [0.0, -1.0, 0.0, 0.0],
        [0.0, 0.0, -1.0, 0.0],
        [0.0, 0.0, 0.0, -1.0],
    ]
)


@pytest.mark.parametrize(
    "mass,spin", [(0.0, -1), (0.0, 1), (1.0, -1.0), (1.0, 0.0), (1.0, 1.0)]
)
def test_vector_wf_transverse(mass: float, spin: int):
    """Test that polarization vectors are transverse.

    Recall that the polarization vectors satisfy:
        p.eps(p) = 0
    """
    prefix = "massless_" if mass == 0.0 else "massive_"
    momenta = np.array(TEST_DATA[prefix + "momenta"])
    helax_wfs: np.ndarray = np.transpose(
        vector_wf(momenta.T, mass, spin, False).wavefunction
    )

    for p, hwf in zip(momenta, helax_wfs):
        rhs = p[0] * hwf[0]
        lhs = p[1] * hwf[1] + p[2] * hwf[2] + p[3] * hwf[3]
        assert rhs == pytest.approx(lhs)


@pytest.mark.parametrize(
    "mass,spin", [(0.0, -1), (0.0, 1), (1.0, -1.0), (1.0, 0.0), (1.0, 1.0)]
)
def test_vector_wf_normalization(mass: float, spin: int):
    """Test the normalization of polarization vectors.

    Recall that polarization vectors are normalized to:
        sum_mu { eps(p, lam)[mu] conj(eps(p, lam)[mu]) } = -1
    """
    prefix = "massless_" if mass == 0.0 else "massive_"
    momenta = np.array(TEST_DATA[prefix + "momenta"])
    helax_wfs: np.ndarray = np.transpose(
        vector_wf(momenta.T, mass, spin, False).wavefunction
    )

    for hwf in helax_wfs:
        norm_sqr = np.abs(hwf) ** 2
        norm = norm_sqr[0] - norm_sqr[1] - norm_sqr[2] - norm_sqr[3]
        assert norm == pytest.approx(-1.0)


@pytest.mark.parametrize("mass", [0.0, 1.0])
def test_vector_wf_completeness(mass: float):
    """Test the completeness relation for polarization vectors.

    Recall that the completeness relation for polarization vectors is:
        sum(lam)(eps(p, lam)[mu] conj(eps(p, lam))) = -g[mu,nu] + p[mu]p[nu]/m^2
    """
    if mass == 0:
        spins = (-1, 1)
        momenta = np.array(TEST_DATA["massless_momenta"])
    else:
        spins = (-1, 0, 1)
        momenta = np.array(TEST_DATA["massive_momenta"])

    helax_wfs: np.ndarray = np.transpose(
        np.array(
            [vector_wf(momenta.T, mass, spin, False).wavefunction for spin in spins]
        ),
        # (spin, lor, p) => (p, spin, lor)
        (2, 0, 1),
    )

    for p, wfs in zip(momenta, helax_wfs):
        expected = -MetricG
        if mass == 0.0:
            pbar = MetricG @ p
            expected += (np.outer(p, pbar) + np.outer(pbar, p)) / ldot(p, pbar)
        else:
            expected += np.outer(p, p) / mass**2

        actual = np.zeros_like(expected, dtype=np.complex128)
        for wf in wfs:
            actual += np.outer(wf, np.conj(wf))

        residual = np.max(np.abs(expected - actual))

        assert residual < 1e-10


def test_vector_wf_massless():
    momenta = np.array(TEST_DATA["massless_momenta"])
    spin_up: np.ndarray = TEST_DATA["massless_up"]
    spin_down: np.ndarray = TEST_DATA["massless_down"]

    helax_spin_up: np.ndarray = np.transpose(
        vector_wf(momenta.T, 0.0, 1, False).wavefunction
    )
    helax_spin_down = np.transpose(vector_wf(momenta.T, 0.0, -1, False).wavefunction)

    for t, h in zip(spin_up, helax_spin_up):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)

    for t, h in zip(spin_down, helax_spin_down):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)


def test_vector_wf_massive():
    momenta = np.array(TEST_DATA["massive_momenta"])
    spin_up: np.ndarray = TEST_DATA["massive_up"]
    spin_zero: np.ndarray = TEST_DATA["massive_zero"]
    spin_down: np.ndarray = TEST_DATA["massive_down"]
    mass = 1.0

    helax_spin_up: np.ndarray = np.transpose(
        vector_wf(momenta.T, mass, 1, False).wavefunction
    )
    helax_spin_zero: np.ndarray = np.transpose(
        vector_wf(momenta.T, mass, 0, False).wavefunction
    )
    helax_spin_down = np.transpose(vector_wf(momenta.T, mass, -1, False).wavefunction)

    for t, h in zip(spin_up, helax_spin_up):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)

    for t, h in zip(spin_zero, helax_spin_zero):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)

    for t, h in zip(spin_down, helax_spin_down):
        for i in range(4):
            assert np.real(h[i]) == approx(np.real(t[i]), rel=1e-4, abs=0.0)
            assert np.imag(h[i]) == approx(np.imag(t[i]), rel=1e-4, abs=0.0)



import math
from typing import Callable, Optional

import chex
import jax
import jax.numpy as jnp
import numpy as np

from .utils import kallen_lambda

SquaredMatrixElement = Callable[[chex.Array], chex.Numeric]


def _rambo_init(key: chex.Array, n: int, batch_size: int) -> chex.Array:
    keys = jax.random.split(key, 4)
    rho1 = jax.random.uniform(keys[0], shape=(n, batch_size))
    rho2 = jax.random.uniform(keys[1], shape=(n, batch_size))
    rho3 = jax.random.uniform(keys[2], shape=(n, batch_size))
    rho4 = jax.random.uniform(keys[3], shape=(n, batch_size))

    ctheta = 2 * rho1 - 1.0
    stheta = jnp.sqrt(1.0 - ctheta**2)
    phi = 2.0 * jnp.pi * rho2
    e = -jnp.log(rho3 * rho4)

    return jnp.array(
        [e, e * stheta * jnp.cos(phi), e * stheta * jnp.sin(phi), e * ctheta]
    )


def _rambo_boost(momenta: chex.Array, cme: float):
    # expecting an array with dimension (4, # final-state-particles)
    # chex.assert_rank(momenta, 2)

    sum_ps = jnp.sum(momenta, axis=1)
    inv_mass = jnp.sqrt(
        sum_ps[0] ** 2 - sum_ps[1] ** 2 - sum_ps[2] ** 2 - sum_ps[3] ** 2
    )
    inv_mass = 1.0 / inv_mass

    bx = -inv_mass * sum_ps[1]
    by = -inv_mass * sum_ps[2]
    bz = -inv_mass * sum_ps[3]

    x = cme * inv_mass
    g = sum_ps[0] * inv_mass
    a = 1.0 / (1.0 + g)

    bdotp = bx * momenta[1] + by * momenta[2] + bz * momenta[3]
    fact = a * bdotp + momenta[0]

    return jnp.array(
        [
            x * (g * momenta[0] + bdotp),
            x * (fact * bx + momenta[1]),
            x * (fact * by + momenta[2]),
            x * (fact * bz + momenta[3]),
        ]
    )


def _rambo_compute_scale_factor(
    momenta: chex.Array, cme: float, masses: chex.Array, iterations: int
) -> chex.Numeric:
    e = momenta[0]

    xi0 = jnp.sqrt(1.0 - jnp.square(jnp.sum(masses) / cme))

    xi = xi0 * jnp.ones_like(e[0])
    for _ in range(iterations):
        deltaf = jnp.hypot(e * xi, masses)
        f = jnp.sum(deltaf, axis=0) - cme
        df = jnp.sum(xi * jnp.square(e) / deltaf, axis=0)
        xi = xi - f / df

    return xi


def _rambo_correct_masses(
    momenta: chex.Array, cme: float, masses: chex.Array, iterations: int
) -> chex.Array:
    xi = _rambo_compute_scale_factor(momenta, cme, masses, iterations)
    return jnp.array(
        [
            jnp.hypot(xi * momenta[0], masses),
            xi * momenta[1],
            xi * momenta[2],
            xi * momenta[3],
        ]
    )


def _rambo_weight_rescale_factors(
    momenta: chex.Array, cme: float, n: int
) -> chex.Numeric:
    modsqr = jnp.sum(jnp.square(momenta[1:]), axis=0)
    mod = jnp.sqrt(modsqr)

    t1 = jnp.sum(mod / cme, axis=0) ** (2 * n - 3)
    t2 = jnp.sum(modsqr / momenta[0], axis=0)
    t3 = jnp.prod(mod / momenta[0], axis=0)

    return t1 / t2 * t3 * cme


def _rambo_compute_weights(momenta: chex.Array, cme: float, n: int, base_wgt: float):
    return _rambo_weight_rescale_factors(momenta, cme, n) * base_wgt


def _rambo_base_weight(n: int, cme: float):
    pi = np.pi
    fact_nm2 = math.factorial(n - 2)
    fact = 1.0 / ((n - 1) * fact_nm2**2)
    return fact * (0.5 * pi) ** (n - 1) * cme ** (2 * n - 4) * (0.5 / pi) ** (3 * n - 4)


def _rambo_generate(
    cme: float,
    masses: chex.Array,
    key: chex.Array,
    batch_size: int,
    iterations: int,
    base_wgt: float,
):
    n = len(masses)
    momenta = _rambo_init(key, n, batch_size)
    momenta = _rambo_boost(momenta, cme)
    momenta = _rambo_correct_masses(momenta, cme, masses, iterations)
    weights = _rambo_compute_weights(momenta, cme, n, base_wgt)
    return momenta, weights


def generate_phase_space(
    cme: float,
    masses: np.ndarray,
    key: chex.Array,
    *,
    msqrd: Optional[SquaredMatrixElement] = None,
    batch_size: int = 10_000,
    jit=True,
):
    n: int = len(masses)
    base_wgt = _rambo_base_weight(n, cme)
    ms = jnp.expand_dims(jnp.array(masses), -1)
    iterations = 10

    if jit:
        generator = jax.jit(_rambo_generate, static_argnums=(3, 4, 5))
    else:
        generator = _rambo_generate

    momenta, weights = generator(cme, ms, key, batch_size, iterations, base_wgt)

    if msqrd is not None:
        weights = msqrd(momenta) * weights

    return momenta, weights


def integrate_phase_space(
    cme: float,
    masses: np.ndarray,
    key: chex.Array,
    *,
    msqrd: Optional[SquaredMatrixElement] = None,
    batch_size: int = 10_000,
    jit=True,
):
    assert cme > sum(
        masses
    ), "Center of mass energy must be greater than sum of `masses`."

    _, weights = generate_phase_space(
        cme, masses, key, msqrd=msqrd, batch_size=batch_size, jit=jit
    )

    integral = jnp.nanmean(weights)
    error = jnp.nanstd(weights, ddof=1) / math.sqrt(batch_size)

    return integral, error


def decay_width(
    mass: float,
    masses: np.ndarray,
    key: chex.Array,
    *,
    msqrd: Optional[SquaredMatrixElement] = None,
    batch_size: int = 10_000,
    jit=True,
):
    assert mass > sum(
        masses
    ), "Center of mass energy must be greater than sum of `masses`."

    integral, error = integrate_phase_space(
        mass, masses, key, msqrd=msqrd, batch_size=batch_size, jit=jit
    )

    pre = 0.5 / mass
    width = pre * integral
    error = pre * error

    return width, error


def cross_section(
    cme: float,
    m1: float,
    m2: float,
    masses: np.ndarray,
    key: chex.Array,
    *,
    msqrd: Optional[SquaredMatrixElement] = None,
    batch_size: int = 10_000,
    jit=True,
):
    assert (
        cme > m1 + m2
    ), "Center of mass energy must be greater than sum of `m1` and `m2`."
    assert cme > sum(
        masses
    ), "Center of mass energy must be greater than sum of `masses`."

    integral, error = integrate_phase_space(
        cme, masses, key, msqrd=msqrd, batch_size=batch_size, jit=jit
    )

    pre = 0.5 / math.sqrt(kallen_lambda(cme**2, m1**2, m2**2))
    width = pre * integral
    error = pre * error

    return width, error



"""Module for computing currents and amplitudes."""


import numpy as np

from helax.numpy.wavefunctions import VectorWf
from helax.numpy.wavefunctions.weyl import WeylType, WeylWf
from helax.vertices import VertexWWV

from .propagators import attach_vector

IM = 1.0j
DAGGER_TYPE = WeylType.Xd | WeylType.Yd


# =============================================================================
# ---- Currents ---------------------------------------------------------------
# =============================================================================


def current_ww_to_v(
    vertex: VertexWWV, mass: float, width: float, chi: WeylWf, eta: WeylWf
) -> VectorWf:
    """Fuse two weyl spinors into a vector boson.

    Parameters
    ----------
    vertex : VertexWWV
        Feynman rule for the W-W-V vertex.
    mass : float
        Mass of the resulting vector.
    width : float
        Width of the resulting vector.
    chi, eta : WeylWf
        Weyl spinors. One must be daggered and the other un-daggered.

    Returns
    -------
    eps: VectorWf
        Resulting vector boson wavefuction.
    """
    dagger_type = WeylType.Xd | WeylType.Yd
    is_chi_dagger = chi.type & dagger_type
    is_eta_dagger = eta.type & dagger_type
    assert is_chi_dagger != is_eta_dagger, (
        "Invalid input spinors."
        + " One spinor must be daggered and the other un-daggered."
    )

    if is_chi_dagger:
        wl = chi.wavefunction
        pout = chi.momentum
        wr = eta.wavefunction
        pin = eta.momentum
    else:
        wl = eta.wavefunction
        pout = eta.momentum
        wr = chi.wavefunction
        pin = chi.momentum

    momentum = pin - pout

    wavefunction = vertex.g * np.array(
        [
            # sigma[0]
            wl[0] * wr[0] + wl[1] * wr[1],
            # sigma[1]
            -(wl[0] * wr[1] + wl[1] * wr[0]),
            # sigma[2]
            -IM * (-wl[0] * wr[1] + wl[1] * wr[0]),
            # sigma[3]
            -(wl[0] * wr[0] - wl[1] * wr[1]),
        ]
    )

    return attach_vector(
        VectorWf(wavefunction=wavefunction, momentum=momentum, direction=1), mass, width
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['helax',
 'helax.jax',
 'helax.jax.testdata',
 'helax.numpy',
 'helax.numpy.amplitudes',
 'helax.numpy.wavefunctions']

package_data = \
{'': ['*'], 'helax.numpy.wavefunctions': ['testdata/*']}

install_requires = \
['numpy>=1.23.3,<2.0.0', 'scipy>=1.9.1,<2.0.0']

extras_require = \
{'gpu': ['jax>=0.3.17,<0.4.0', 'jaxlib>=0.3.15,<0.4.0', 'chex>=0.1.5,<0.2.0']}

setup_kwargs = {
    'name': 'helax',
    'version': '0.1.3',
    'description': 'Python package for computing helicity amplitudes',
    'long_description': '# `helax`: Helicity Amplitudes with Jax or NumPy\n\nImplementation from scratch of the HELAS library using Jax and NumPy.\n',
    'author': 'Logan A. Morrison',
    'author_email': 'loganmorrison99@gmail.com',
    'maintainer': 'Logan Morrison',
    'maintainer_email': 'loganmorrison99@gmail.com',
    'url': 'https://github.com/LoganAMorrison/helax',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<3.12',
}


setup(**setup_kwargs)



import jax
import jax.numpy as jnp


def kallen_lambda(a, b, c):
    return a**2 + b**2 + c**2 - 2 * a * b - 2 * a * c - 2 * b * c


@jax.jit
def abs2(arr: jnp.ndarray):
    return jnp.real(arr) ** 2 + jnp.imag(arr) ** 2



import numpy as np

from .typing import NumericArray


def kallen_lambda(a, b, c):
    return a**2 + b**2 + c**2 - 2 * a * b - 2 * a * c - 2 * b * c


def abs2(arr: NumericArray):
    return np.real(arr) ** 2 + np.imag(arr) ** 2



"""Module for computing currents and amplitudes.


Notes on signs
==============

All momentum are considered incoming, except for fermions. For fermions, the
momentum always points along the direction of the fermion flow.

[v,s](p) + [v,s](q) -> [v,s](k)
-----------------------
k = p + q

    p ->     k=p+q ->
    -------x------
          /
    -----
    q ->


f(p) + [v,s](q) -> f(k)
-----------------------
k = p + q

    p ->     ->  k=p+q
    --->---x--->---
          /
    -----
    q ->


fbar(p) + [v,s](q) -> fbar(k)
-----------------------------
k = p - q

    p <-     <- k=p-q
    ---<---x---<---
          /
    -----
    q ->


f(p) + fbar(q) -> [v,s](k)
--------------------------
k = p - q

    p ->     -> k=p-q
    --->---x------
          /
    --<--
    q <-


w(p) + w(q) -> s(k)
-------------------
k = p + q

    p ->     -> k=p+q
    --->---x------
          /
    -->--
    q ->


w^+(p) + w^+(q) -> s(k)
---------------------------
k = -p - q

    p <-     -> k=-p-q
    ---<---x------
          /
    --<--
    q <-
"""


import numpy as np

from helax.numpy.dirac import WeylS0, WeylS1, WeylS2, WeylS3
from helax.numpy.lvector import ldot, lnorm_sqr
from helax.numpy.typing import RealArray
from helax.numpy.wavefunctions import DiracWf, ScalarWf, VectorWf
from helax.numpy.wavefunctions.weyl import WeylType, WeylWf

IM = 1.0j
DAGGER_TYPE = WeylType.Xd | WeylType.Yd


def complex_mass_sqr(mass: float, width: float) -> complex:
    """Returns the complexified mass given the particle's width."""
    return mass * (mass - 1j * width)


def propagator_den(momentum: RealArray, mass: float, width: float) -> RealArray:
    """Returns the scalar component of the propagator."""
    return IM / (lnorm_sqr(momentum) - complex_mass_sqr(mass, width))


def attach_dirac(psi: DiracWf, mass: float, width: float) -> DiracWf:
    """Attach a propagator to a Dirac wavefunction."""
    p = psi.momentum
    f = psi.wavefunction

    den = propagator_den(psi.momentum, mass, width)

    p1p2 = p[1] + IM * p[2]
    p1m2 = p[1] - IM * p[2]
    p0p3 = p[0] + p[3]
    p0m3 = p[0] - p[3]

    if psi.direction == -1:
        wavefunction = np.array(
            [
                (mass * f[0] - f[3] * p1m2 + f[2] * p0m3) * den,
                (mass * f[1] - f[2] * p1p2 + f[3] * p0p3) * den,
                (mass * f[2] + f[1] * p1m2 + f[0] * p0p3) * den,
                (mass * f[3] + f[0] * p1p2 + f[1] * p0m3) * den,
            ]
        )
    else:
        wavefunction = np.array(
            [
                (mass * f[0] + f[3] * p1p2 + f[2] * p0p3) * den,
                (mass * f[1] + f[2] * p1m2 + f[3] * p0m3) * den,
                (mass * f[2] - f[1] * p1p2 + f[0] * p0m3) * den,
                (mass * f[3] - f[0] * p1m2 + f[1] * p0p3) * den,
            ]
        )

    return DiracWf(
        wavefunction=wavefunction, momentum=psi.momentum, direction=psi.direction
    )


def attach_vector(eps: VectorWf, mass: float, width: float) -> VectorWf:
    wf = eps.wavefunction
    k = eps.momentum
    den = propagator_den(k, mass, width)

    if mass == 0:
        invcm2 = 0.0j
    else:
        invcm2 = 1.0 / complex_mass_sqr(mass, width)

    wf = (-wf + k * ldot(k, wf) * invcm2) * den

    return VectorWf(wavefunction=wf, momentum=k, direction=eps.direction)


def attach_scalar(phi: ScalarWf, mass: float, width: float) -> ScalarWf:
    wf = phi.wavefunction * propagator_den(phi.momentum, mass, width)
    return ScalarWf(wavefunction=wf, momentum=phi.momentum, direction=phi.direction)


def attach_weyl(chi: WeylWf, mass: float, width: float) -> WeylWf:
    """Attach a propagator to a Weyl wavefunction."""
    p = chi.momentum
    chi1 = chi.wavefunction[0]
    chi2 = chi.wavefunction[1]

    den = propagator_den(p, mass, width)

    if chi.type & DAGGER_TYPE:
        p_dot_sigma = p[0] * WeylS0 + p[1] * WeylS1 + p[2] * WeylS2 + p[3] * WeylS3
        # If the spinor is a daggered spinor, then we want to compute
        # z^b = x_a sigma^ab. This will yield a spinor with a raised index.
        # To lower the index, we contract with the spinor metric tensor:
        #   z_a = eps_ab * z^b
        #     => z_1 = eps_12 * z^2 = -z^2
        #     => z_2 = eps_21 * z^1 = +z^1
        eta1 = IM * (p_dot_sigma[0, 0] * chi1 + p_dot_sigma[0, 1] * chi2) * den
        eta2 = IM * (p_dot_sigma[1, 0] * chi1 + p_dot_sigma[1, 1] * chi2) * den

        eta1, eta2 = -eta2, eta1

    else:
        p_dot_sigma = p[0] * WeylS0 - p[1] * WeylS1 - p[2] * WeylS2 - p[3] * WeylS3
        # If the spinor isn't a daggered spinor, then we want to compute
        # z_b = x^a sigma_ab. But our spinor is x_a. Need to raise the index,
        # yielding: x^a = eps^ab x_b
        #    => x^1 = eps^12 x_2 = +x_2
        #    => x^2 = eps^21 x_1 = -x_1
        chi1, chi2 = chi2, -chi1

        eta1 = IM * (p_dot_sigma[0, 0] * chi1 + p_dot_sigma[0, 1] * chi2) * den
        eta2 = IM * (p_dot_sigma[1, 0] * chi1 + p_dot_sigma[1, 1] * chi2) * den

    wavefunction = np.array([eta1, eta2])

    if chi.type == WeylType.X:
        new_type = WeylType.Xd
    elif chi.type == WeylType.Xd:
        new_type = WeylType.X
    elif chi.type == WeylType.Y:
        new_type = WeylType.Yd
    elif chi.type == WeylType.Yd:
        new_type = WeylType.Y
    else:
        raise ValueError(f"Invalid weyl type '{chi.type}'.")

    return WeylWf(wavefunction=wavefunction, momentum=p, type=new_type)


def attach_weyl_mass(chi: WeylWf, mass: float, width: float) -> WeylWf:
    """Attach a propagator to a Dirac wavefunction."""
    p = chi.momentum
    wf = chi.wavefunction

    den = propagator_den(p, mass, width)
    prop = IM * mass * den
    wavefunction = np.array([prop * wf[0], prop * wf[1]])

    return WeylWf(wavefunction=wavefunction, momentum=p, type=chi.type)



from typing import Union

import numpy as np
import numpy.typing as npt

Numeric = Union[np.uint, np.float_, np.complex_, np.int_, np.uint]
NdArray = npt.NDArray[Numeric]


def ldot(lv1: NdArray, lv2: NdArray):
    return lv1[0] * lv2[0] - lv1[1] * lv2[1] - lv1[2] * lv2[2] - lv1[3] * lv2[3]


def lnorm_sqr(lv: NdArray):
    return ldot(lv, lv)


def lnorm(lv: NdArray):
    return np.sqrt(lnorm_sqr(lv))


def lnorm3_sqr(lv: NdArray):
    return np.square(lv[1:])


def lnorm3(lv: NdArray):
    return np.square(lnorm3_sqr(lv))



import dataclasses

import numpy as np

from helax.numpy.typing import ComplexArray, RealArray

from .utils import check_momentum_shape


@dataclasses.dataclass
class VectorWf:
    wavefunction: ComplexArray
    momentum: RealArray
    direction: int


# =============================================================================
# ---- Vector Wavefunction ----------------------------------------------------
# =============================================================================


def __check_spin_vector(*, spin: int):
    assert spin == 1 or spin == -1 or spin == 0, "Spin must be 1, 0, or -1."


def __polvec_transverse(*, k: RealArray, spin: int, sgn: int) -> RealArray:
    """
    Compute a transverse (spin 1 or -1) vector wavefunction.

    Parameters
    ----------
    k: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    spin: int
        Spin of the particle. Must be -1, 0, or -1.
    s: int
        If 1, the returned wavefunction is outgoing. If -1, the
        returned wavefunction is incoming. Must be 1 or -1.
    """
    assert sgn == 1 or sgn == -1, "`s` value must be 1 or -1."
    check_momentum_shape(k, require_batch=True)
    kx, ky, kz = k[1:]
    kt = np.hypot(kx, ky)

    eps = np.finfo(k.dtype.name).eps
    mask = kt < eps
    polvec = np.zeros((4, kx.shape[-1]), dtype=kx.dtype) * 1j

    if np.any(mask):
        polvec[0, mask] = 0.0j
        polvec[1, mask] = -spin / np.sqrt(2) + 0.0j
        polvec[2, mask] = -np.copysign(1.0, kz[mask]) * 1.0j / np.sqrt(2)
        polvec[3, mask] = 0.0j

    mask = ~mask
    if np.any(mask):
        kx = kx[mask]
        ky = ky[mask]
        kz = kz[mask]
        km = np.sqrt(np.square(kx) + np.square(ky) + np.square(kz))

        kxt = kx / kt[mask] / np.sqrt(2)
        kyt = ky / kt[mask] / np.sqrt(2)
        kzm = kz / km
        ktm = kt[mask] / km / np.sqrt(2)

        polvec[0, mask] = 0.0 + 0.0 * 1j
        polvec[1, mask] = -spin * kxt * kzm + +sgn * kyt * 1j
        polvec[2, mask] = -spin * kyt * kzm + -sgn * kxt * 1j
        polvec[3, mask] = +spin * ktm + 0.0 * 1j

    return polvec


def __polvec_longitudinal(*, k: RealArray, mass: float) -> RealArray:
    """
    Compute a longitudinal (spin 0) vector wavefunction.

    Parameters
    ----------
    k: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    """
    check_momentum_shape(k, require_batch=True)
    polvec = np.zeros((4, k.shape[-1]), dtype=k.dtype) * 1j
    if mass == 0.0:
        return polvec

    e, kx, ky, kz = k
    km = np.linalg.norm(k[1:], axis=0)
    eps = np.finfo(k.dtype.name).eps
    mask = km < eps

    if np.any(mask):
        polvec[3, mask] = 1.0 + 0.0j

    mask = ~mask
    if np.any(mask):
        n = e[mask] / (mass * km[mask])
        polvec[0, mask] = km[mask] / mass + 0.0j
        polvec[1, mask] = n * kx[mask] + 0.0j
        polvec[2, mask] = n * ky[mask] + 0.0j
        polvec[3, mask] = n * kz[mask] + 0.0j

    return polvec


def __vector_wf(*, momentum: RealArray, mass: float, spin: int, sgn: int):
    """
    Compute a vector wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be -1, 0, or -1.
    s: int
        If 1, the returned wavefunction is outgoing. If -1, the
        returned wavefunction is incoming. Must be 1 or -1.
    """
    assert sgn == 1 or sgn == -1, "`s` value must be 1 (incoming) or -1 (outgoing)."
    check_momentum_shape(momentum, require_batch=True)

    if spin == -1 or spin == 1:
        return __polvec_transverse(k=momentum, spin=spin, sgn=sgn)
    return __polvec_longitudinal(k=momentum, mass=mass)


def vector_wf(momentum: RealArray, mass: float, spin: int, out: bool) -> VectorWf:
    """
    Compute a vector wavefunction.

    Parameters
    ----------
    momentum: ndarray
        Array containing the four-momentum of the particle.
        Must be 1 or 2 dimensional with leading dimension of size 4.
    mass: float
        Mass of the particle.
    spin: int
        Spin of the particle. Must be -1, 0, or -1.
    out: bool
        If true, the returned wavefunction is outgoing.
    """
    __check_spin_vector(spin=spin)
    check_momentum_shape(momentum, require_batch=False)
    sgn = -1 if out else 1

    if len(momentum.shape) == 2:
        k = momentum
        squeeze = False
    else:
        k = np.expand_dims(momentum, -1)
        squeeze = True

    wf = __vector_wf(momentum=k, mass=mass, spin=spin, sgn=sgn)

    if squeeze:
        wf = np.squeeze(wf)

    return VectorWf(wavefunction=wf, momentum=sgn * momentum, direction=sgn)
